/*
 * commands.c
 * functions for powwow's built-in #commands
 *
 * (created: Finn Arne Gangstad (Ilie), Dec 25th, 1993)
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <sys/time.h>
#include <errno.h>
#include <assert.h>

#ifdef USE_REGEXP
#  include <regex.h>
#endif

#include "powwow.h"
#include "beam.h"
#include "commands.h"
#include "edit.h"
#include "list.h"
#include "mapping.h"
#include "tcp.h"
#include "tty.h"
#include "eval.h"

void show_stat();
static void show_aliases();
static void show_actions();
static void show_prompts();
static void show_marks();
static void show_keybinds();
void show_delaynode();
static void parse_bind();
static void define_new_key();
static void print_seq();
static int get_one_char();
static void delete_action();
static void delete_prompt();
static void change_actionorprompt();
static void add_new_action();
static void add_new_prompt();
static void show_attr_syntax();
static void show_history();
static void exe_history();
static char *redirect();

static void cmd_help(), cmd_shell(),
  cmd_action(), cmd_add(), cmd_alias(), cmd_at(), cmd_beep(), cmd_bind(),
  cmd_cancel(), cmd_capture(), cmd_connect(), cmd_cpu(),
  cmd_do(), cmd_delim(), cmd_edit(), cmd_emulate(), cmd_exe(),
  cmd_file(), cmd_for(), cmd_hilite(), cmd_history(), cmd_host(),
  cmd_identify(), cmd_if(), cmd_in(), cmd_init(), cmd_isprompt(), cmd_keyexe(),
  cmd_lines(), cmd_load(), cmd_map(), cmd_mark(), cmd_net(), cmd_nice(),
  cmd_option(), cmd_prefix(), cmd_print(), cmd_prompt(), cmd_put(),
  cmd_qui(), cmd_quit(), cmd_quote(),
  cmd_rebind(), cmd_rebindall(), cmd_rebindALL(),
  cmd_record(), cmd_reset(), cmd_retrace(),
  cmd_save(), cmd_send(), cmd_settimer(), cmd_snoop(), cmd_spawn(), cmd_stop(),
  cmd_time(), cmd_var(), cmd_ver(), cmd_while(), cmd_write(),
  cmd_eval(), cmd_zap();

#ifdef BUG_TELNET
static void cmd_color();
#endif

cmdstruct commands[] =
{
    {"help", "[keys|math|command]\tthis text, or help on specified topic",
	  cmd_help},
    {"17", "command\t\t\trepeat `command' 17 times", (void (*)())NULL},
#ifndef NO_SHELL
    {"!", "shell-command\t\texecute a shell command using /bin/sh", cmd_shell},
#endif
    {"action", "[[<|=|>|%][+|-]name] [{pattern|(expression)} [=[action]]]\n\t\t\t\tdelete/list/define actions",
	  cmd_action},
    {"add", "{string|(expr)}\t\tadd the string to word completion list", 
	  cmd_add},
    {"alias", "[name[=[text]]]\t\tdelete/list/define aliases", cmd_alias},
    {"at", "[name [(time-string) [command]]\tset time of delayed label",
	  cmd_at},
    {"beep", "\t\t\t\tmake your terminal beep (like #print (*7))",
	  cmd_beep},
    {"bind", "[edit|name [seq][=[command]]]\tdelete/list/define key bindings",
	  cmd_bind},
    {"cancel", "[number]\t\tcancel editing session", cmd_cancel},
    {"capture", "[filename]\t\tbegin/end of capture to file", cmd_capture},
#ifdef BUG_TELNET
    {"color", "attr\t\t\tset default colors/attributes", cmd_color},
#endif
    {"connect", "[connect-id [initstr] [address port]\topen a new connection",
	  cmd_connect},
    {"cpu", "\t\t\t\tshow CPU time used by powwow", cmd_cpu},
    {"delim", "[normal|program|{custom [chars]}]\n\t\t\t\tset word completion delimeters",
          cmd_delim},
    {"do", "(expr) command\t\trepeat `command' (expr) times", cmd_do},
    {"edit", "\t\t\t\tlist editing sessions", cmd_edit},
    {"emulate", "[<|!]{text|(expr)}\tprocess result as if received from host", cmd_emulate},
    {"exe", "[<|!]{text|(string-expr)}\texecute result as if typed from keyboard", cmd_exe},
    {"file", "[=[filename]]\t\tset/show powwow definition file", cmd_file},
    {"for", "([init];check;[loop]) command\twhile `check' is true exec `command'", cmd_for},
    {"hilite", "[attr]\t\t\thighlight your input line", cmd_hilite},
    {"history", "[{number|(expr)}]\tlist/execute commands in history",
	  cmd_history},
    {"host", "[hostname port]]\tset/show address of default host", cmd_host},
    {"identify", "[startact [endact]]\tsend MUME client identification",
	  cmd_identify},
    {"if", "(expr) instr1 [; #else instr2]\tif `expr' is true execute `instr1'\n\t\t\t\totherwise execute `instr2'", cmd_if},
    {"in", "[label [(delay) [command]]]\tdelete/list/define delayed labels",
	  cmd_in},
    {"init", "[=[command]]\t\tdefine command to execute on connect to host",
	  cmd_init},
    {"isprompt", "\t\t\trecognize a prompt as such", cmd_isprompt},
    {"key", "name\t\t\texecute the `name' key binding", cmd_keyexe},
    {"lines", "[number]\t\t\tshow/set terminal lines", cmd_lines},
    {"load", "[filename]\t\tload powwow settings from file", cmd_load},
    {"map", "[-[number]|walksequence]\tshow/clear/edit (auto)map", cmd_map},
    {"mark", "[string[=[attr]]]\t\tdelete/list/define markers", cmd_mark},
    {"net", "\t\t\t\tprint amount of data received from/sent to host", cmd_net},
    {"nice", "[{number|(expr)}[command]]\tset/show priority of new actions/marks", cmd_nice},
    {"option", "[[+|-|=]name]\t\tturn various options", cmd_option},
    {"prefix", "string\t\t\tprefix all lines with string", cmd_prefix},
    {"", "(expr)\t\t\tevaluate expression, trashing result", cmd_eval},
    {"print", "[<|!][text|(expr)]\tprint text/result on screen, appending a \\n\n\t\t\t\tif no argument, prints value of variable $0", 
	  cmd_print},
    {"prompt", "[[<|=|>|%][+|-]name] [{pattern|(expression)} [=[prompt-action]]]\n\t\t\t\tdelete/list/define actions on prompts",
	  cmd_prompt},
    {"put", "{text|(expr)}\t\tput text/result of expression in history",
	  cmd_put},
    {"qui", "\t\t\t\tdo nothing", cmd_qui},
    {"quit", "\t\t\t\tquit powwow", cmd_quit},
    {"quote", "[on|off]\t\t\ttoggle verbatim-flag on/off", cmd_quote},
    {"rebind", "name [seq]\t\tchange sequence of a key binding",
	  cmd_rebind},
    {"rebindall", "\t\t\trebind all key bindings",
	  cmd_rebindall},
    {"rebindALL", "\t\t\trebind ALL key bindings, even trivial ones",
	  cmd_rebindALL},
    {"record", "[filename]\t\tbegin/end of record to file", cmd_record},
    {"reset", "<list-name>\t\tclear the whole defined list and reload default",
	  cmd_reset},
    {"retrace", "[number]\t\tretrace the last number steps", cmd_retrace},
    {"save", "[filename]\t\tsave powwow settings to file", cmd_save},
    {"send", "[<|!]{text|(expr)}\teval expression, sending result to the MUD", cmd_send},
    {"settimer", "(expr)\t\tsets internal timer to expr milliseconds",
	  cmd_settimer},
    {"snoop", "connect-id\t\ttoggle output display for connections", cmd_snoop},
    {"spawn", "connect-id command\ttalk with a shell command", cmd_spawn},
    {"stop", "\t\t\t\tremove all delayed commands from active list", cmd_stop},
    {"time", "\t\t\t\tprint current time and date", cmd_time},
    {"var", "variable [= [<|!]{string|(expr)} ]\twrite result into the variable", cmd_var},
    {"ver", "\t\t\t\tshow powwow version", cmd_ver},
    {"while", "(expr) instr\t\twhile `expr' is true execute `instr'",
	  cmd_while},
    {"write", "[>|!](expr;name)\t\twrite result of expr to `name' file",
	  cmd_write},
    {"zap", "connect-id\t\t\tclose a connection", cmd_zap},
    {(char*)NULL, (char*)NULL, (void (*)())NULL}
};

/* anyone knows if ANSI 6429 talks about more than 8 colors? */
static char *colornames[] = {
    "black", "red", "green", "yellow", "blue", "magenta", "cyan", "white",
    "BLACK", "RED", "GREEN", "YELLOW", "BLUE", "MAGENTA", "CYAN", "WHITE", "none"
};

static void cmd_nice(arg)
char *arg;
{
    int nnice = a_nice;
    arg = skipspace(arg);
    if (!*arg) {
	PRINT_INT ("#nice: %d\n", a_nice);
	return;
    }
    if (isdigit(*arg)) {
	a_nice = 0;
	while (isdigit(*arg)) {
	    a_nice *= 10;
	    a_nice += *arg++ - '0';
	}
    }
    else if (*arg++=='(') {
	char buf[LINELEN], *tmp = buf;
	int type;

	type = eval(&tmp, &arg, PRINT_AS_NUM);
	if (error)
	  return;
	if (type!=TYPE_NUM) {
	    print_error(error=NO_NUM_VALUE_ERROR);
	    return;
	}    
	if (*arg++ != ')') {
	    PRINT_INT ("#nice: ");
	    print_error(error=MISSING_PAREN_ERROR);
	    return;
	}
	a_nice = (int)*(long *)buf;
	if (a_nice<0)
	  a_nice = 0;
    }
    arg = skipspace(arg);
    if (*arg) {
	parse_instruction(arg, 0, 0);
	a_nice = nnice;
    }
}

static void cmd_init(arg)
char *arg;
{
    arg = skipspace(arg);
    
    if (*arg == '=')
      if (*++arg) {
	  strcpy(initstr, arg);
	  if (echo_int) {
	      PRINT_INT ("#init: %s\n", initstr);
	  }
      }
    else {
	*initstr = '\0';
	if (echo_int) {
	    PRINT_INT ("#init cleared.\n");
	}
    }
    else {
	char buf[LINELEN];
	strcpy(buf, "#init =");
	strcat(buf, initstr);
	to_input_line(buf);
    }
}

static void cmd_host(arg)
char *arg;
{
    char newhost[LINELEN];
    
    arg = skipspace(arg);
    if (*arg) {
	arg = split_first_word(newhost, arg);
	if (*arg) {
	    strcpy(hostname, newhost);
	    strcpy(portnumber, arg);
	    if (echo_int) {
		PRINT_INT("#host set to: %s %s\n", hostname, portnumber);
	    }
	} else {
	    PRINT_INT("#host: missing portnumber.\n");
	}
    } else if (*hostname) {
	sprintf(newhost, "#host %s %s", hostname, portnumber);
	to_input_line(newhost);
    } else {
	PRINT_INT("#syntax: #host hostname port\n");
    }
}

/*ARGSUSED*/
static void cmd_beep(arg)
char *arg;
{
    tty_putc((char)'\007');
}

static void cmd_delim(arg)
char *arg;
{
    char buf[LINELEN];
    int n;

    arg = skipspace(arg);
    if (!*arg) {
	PRINT_INT("#delim: %s (%s)\n", delim_mode_names[word_delim_mode],
		  WORDDELIM);
	return;
    }

    arg = split_first_word(buf, arg);
    n = 0;
    while (n < DELIM_MODES &&
	   strcmp (delim_mode_names[n], buf) != 0)
	n++;
    
    if (n >= DELIM_MODES) {
	PRINT_INT("#delim [normal|program|{custom <chars>}\n");
	return;
    }

    if (n == DELIM_CUSTOM) {
	arg = split_first_word(buf + 1, arg);
	*buf = ' ';				/* force ' ' in the delims */
	set_custom_delimeters(buf);
    }
    word_delim_mode = n;
}

static void cmd_hilite(arg)
char *arg;
{
    int attr;
    
    arg = skipspace(arg);
    attr = parse_attributes(arg);
    if (attr == -1) {
	PRINT_INT ("#attribute syntax error.\n");
	if (echo_int)
	  show_attr_syntax();
    } else {
	attr_string(attr, edattrbeg, edattrend);
	if (echo_int) {
	    PRINT_INT ("#input highlighting is now %so%s%s.\n",
		       edattrbeg, (attr == NOATTRCODE) ? "ff" : "n", edattrend);
	}
    }
}

/*
 * parse attribute description in line.
 * Return attribute if successful, -1 otherwise.
 */
int parse_attributes(line)
char *line;
{
    int strcasecmp();
    char *p;
    int tok = 0, fg, bg, t = -1;
    
    if (!(p = strtok(line, " ")))
      return NOATTRCODE;
    
    fg = bg = NO_COLOR;
    
    while (t && p) {
	if (!strcasecmp(p, "bold"))
	  t = ATTR_BOLD;
	else if (!strcasecmp(p, "blink"))
	  t = ATTR_BLINK;
	else if (!strcasecmp(p, "underline"))
	  t = ATTR_UNDERLINE;
	else if (!strcasecmp(p, "inverse") || !strcasecmp(p, "reverse"))
	  t = ATTR_INVERSE;
	else
	  t = 0;
	
	if (t) {	 
	    tok |= t;
	    p = strtok(NULL, " ");
	}
    }
    
    if (!p)
      return ATTRCODE(tok, fg, bg);
    
    for (t = 0; t <= COLORS && strcmp(p, colornames[t]); t++)
      ;
    if (t <= COLORS) {
	fg = t;
	p = strtok(NULL, " ");
    }
    
    if (!p)
      return ATTRCODE(tok, fg, bg);
    
    if (strcasecmp(p, "on"))
      return -1;      /* invalid attribute */
    
    if (!(p = strtok(NULL, " ")))
      return -1;
    
    for (t = 0; t <= COLORS && strcmp(p, colornames[t]); t++)
      ;
    if (t <= COLORS)
      bg = t;
    else
      return -1;
    
    return ATTRCODE(tok, fg, bg);
}

/*
 * display attribute syntax
 */
static void show_attr_syntax()
{
    int i;
    PRINT_INT ("#attribute syntax:\n\tOne or more of:\tbold, blink, underline, inverse\n");
    tty_puts("\tand/or\t[foreground] [ON background]\n");
    tty_puts("\tColors: ");
    for (i = 0; i < COLORS; i++)
      tty_printf("%s%s", colornames[i],
		 (i == LOWCOLORS - 1 || i == COLORS - 1) ? "\n\t\t" : ",");
    tty_printf("%s\n", colornames[i]);
}

/*
 * return a static pointer to name of given attribute code
 */
char *attr_name(attrcode)
int attrcode;
{
    static char name[LINELEN];
    int fg = FOREGROUND(attrcode), bg = BACKGROUND(attrcode), tok = ATTR(attrcode);
    
    name[0] = 0;
    if (tok > (ATTR_BOLD | ATTR_BLINK | ATTR_UNDERLINE | ATTR_INVERSE) || fg > COLORS || bg > COLORS)
      return name;   /* error! */
    
    if (tok & ATTR_BOLD)
      strcat(name, "bold ");
    if (tok & ATTR_BLINK)
      strcat(name, "blink ");
    if (tok & ATTR_UNDERLINE)
      strcat(name, "underline ");
    if (tok & ATTR_INVERSE)
      strcat(name, "inverse ");
    
    if (fg < COLORS || (fg == bg == COLORS && !tok))
      strcat(name, colornames[fg]);
    
    if (bg < COLORS) {
	strcat(name, " on ");
	strcat(name, colornames[bg]);
    }
    
    if (!*name)
      strcpy(name, "none");
    
    return name;
}

/*
 * put escape sequences to turn on/off an attribute in given buffers
 */
void attr_string(attrcode, begin, end)
int attrcode; char *begin, *end;
{
    int fg = FOREGROUND(attrcode), bg = BACKGROUND(attrcode),
      tok = ATTR(attrcode);
    *begin = *end = '\0';
    
    if (tok > (ATTR_BOLD | ATTR_BLINK | ATTR_UNDERLINE | ATTR_INVERSE) || fg > COLORS
	|| bg > COLORS)
      return;    /* error! */
    
    if (fg < COLORS) {
	if (bg < COLORS)
	  sprintf(begin, "\033[%c%d;%s%dm",
		  fg<LOWCOLORS ? '3' : '9', fg % LOWCOLORS,
		  bg<LOWCOLORS ? "4" :"10", bg % LOWCOLORS);
	else
	  sprintf(begin, "\033[%c%dm",
		  fg<LOWCOLORS ? '3' : '9', fg % LOWCOLORS);
    }
    else if (bg < COLORS)
	sprintf(begin, "\033[%s%dm", 
		bg<LOWCOLORS ? "4" : "10", bg % LOWCOLORS);
    
    if (tok & ATTR_BOLD) {
	if (tty_modebold[0])
	    strcat(begin, tty_modebold);
	else {
	    strcat(begin, tty_modestandon);
	    strcpy(end, tty_modestandoff);
	}
    }
    
    if (tok & ATTR_BLINK) {
	if (tty_modeblink[0])
	    strcat(begin, tty_modeblink);
	else {
	    strcat(begin, tty_modestandon);
	    strcpy(end, tty_modestandoff);
	}
    }
    
    if (tok & ATTR_UNDERLINE) {
	if (tty_modeuline[0])
	    strcat(begin, tty_modeuline);
	else {
	    strcat(begin, tty_modestandon);
	    strcpy(end, tty_modestandoff);
	}
    }
    
    if (tok & ATTR_INVERSE) {
	if (tty_modeinv[0])
	    strcat(begin, tty_modeinv);
	else {
	    strcat(begin, tty_modestandon);
	    strcpy(end, tty_modestandoff);
	}
    }
    
    if (attrcode != NOATTRCODE && !*end)
	strcpy(end, tty_modenorm);
}

static void cmd_mark(arg)
char *arg;
{
    if (!*arg)
	show_marks();
    else
	parse_marker(arg);
}

static void cmd_reset(arg)
char *arg;
{
    char all = 0;
    arg = skipspace(arg);
    if (!*arg) {
        PRINT_INT ("#reset: must specify one of:\n");
        tty_puts(" alias, action, bind, in (or at), mark, prompt, var, all.\n");
	return;
    }
    all = !strcmp(arg, "all");
    if (all || !strcmp(arg, "alias")) {
        while (aliases)
	    delete_aliasnode(&aliases);
	if (!all)
	    return;
    }
    if (all || !strcmp(arg, "action")) {
        while (actions)
	    delete_actionnode(&actions);
	if (!all)
	    return;
    }
    if (all || !strcmp(arg, "bind")) {
        while (keydefs)
	    delete_keynode(&keydefs);
	tty_add_initial_binds();
	tty_add_walk_binds();
	if (!all)
	    return;
    }
    if (all || !strcmp(arg, "in") || !strcmp(arg, "at")) {
        while (delays)
	    delete_delaynode(&delays);
        while (dead_delays)
	    delete_delaynode(&dead_delays);
	if (!all)
	    return;
    }
    if (all || !strcmp(arg, "mark")) {
        while (markers)
	    delete_marknode(&markers);
	if (!all)
	    return;
    }
    if (all || !strcmp(arg, "prompt")) {
        while (prompts)
	    delete_promptnode(&prompts);
	if (!all)
	    return;
    }
    if (all || !strcmp(arg, "var")) {
        int n;
	char *p;
	varnode **first;

        while (named_vars[0])
	    delete_varnode(&named_vars[0], (char)0);
	first = &named_vars[1];
	while (*first) {
	    if (is_permanent_variable(*first))
		first = &(*first)->next;
	    else	
	      delete_varnode(first, (char)1);
	}
	
	for (n = 0; n < NUMVAR; n++) {
	    var[n].num = 0;
	    if ((p = var[n].str)) {
	        free(p);
		var[n].str = NULL;
	    }
	}
	if (!all)
	    return;
    }
}

static void cmd_keyexe(arg)
char *arg;
{
    keynode *p, *q=NULL;
    
    if (!*arg) return;
    
    for (p = keydefs; p && !q; p = p->next)
	if (!strcmp(p->name, arg))
	q = p;
    
    if (q) {
	if (q->funct == key_run_command)
	    parse_instruction(q->call_data, 0, 0);
	else {
	    PRINT_INT("#sorry, #key cannot be used with editing keys.\n");
	}
    }
    else {
	PRINT_INT("#no such key: %s\n", arg);
    }
}

/*
 * turn ASCII description of a sequence
 * into raw escape sequence
 * return pointer to end of ASCII description
 */
static char *unescape_seq(buf, seq, seqlen)
char *buf, *seq; int *seqlen;
{
    char c, *start = buf;
    int escape = 0, caret = 0;
    
    for (; (c = *seq); seq++) {
        if (escape) {
	    if (c != ESC)
		escape = 0;
	    if (c == ESC2)
		*(buf++) = ESC;	    
	    else
		*(buf++) = c;
        } else if (caret) {
	    /*
	     * handle ^@ ^A ... ^_ as expected:
	     * ^@ == 0x00, ^A == 0x01, ... , ^_ == 0x1f
	     */
	    if (c > 0x40 && c < 0x60)
		*(buf++) = c & 0x1f;
	    /* special case: ^? == 0x7f */
	    else if (c == '?')
		*(buf++) = 0x7f;
	    caret = 0;
        } else if (c == ESC) {
	    /*
	     * GH: \012 ==> octal number
	     */
	    if (ISODIGIT(*(seq + 1)) && ISODIGIT(*(seq + 2)) && 
		ISODIGIT(*(seq + 3))) {
		*(buf++) =(((*++seq - '0') << 6) | 
			   ((*++seq - '0') << 3) |
			   (*++seq - '0'));
	    } else 
		escape = 1;
	} else if (c == '^')
	    caret = 1;
        else if (c == '=')
	    break;
        else
	    *(buf++) = *seq;
    }
    *buf = '\0';
    *seqlen = buf - start;
    return seq;
}

/*
 * read a single character from tty, with timeout in microseconds.
 * timeout == 0 means wait indefinitely (no timeout).
 * return char or -1 if timeout was reached.
 */
static int get_one_char(timeout)
int timeout;
{
    int select();
    extern int errno;
    int err;
    fd_set fds;
    char c;
    struct timeval timeoutbuf;
    
    FD_ZERO(&fds);
    FD_SET(tty_read_fd, &fds);
    timeoutbuf.tv_sec = 0;
    timeoutbuf.tv_usec = timeout;
    err = select(tty_read_fd+1, &fds, NULL, NULL,
		 timeout ? &timeoutbuf : NULL);
    if (err == 0 || (err == -1 && errno == EINTR))
	return -1;
    if (err == -1) {
	errmsg("select");
	return -1;
    }
    while ((err = tty_read(tty_read_fd, &c, 1)) < 0 && errno == EINTR)
	;
    if (err != 1) {
	errmsg("read from tty");
	return -1;
    }
    return (int)(unsigned char)c;
}

/*
 * print an escape sequence in human-readably form.
 */
static void print_seq(seq, len)
char *seq; int len;
{
    char ch;

    while (len--) {
	ch = *(seq++);
        if (ch == '\033') {
            tty_puts("esc ");
            continue;
        }
        if (ch >=0 && ch < ' ') {
            tty_putc('^');
            ch |= '@';
        }
        if (ch == ' ')
	    tty_puts("space ");
        else if (ch == 0x7f)
	    tty_puts("del ");
	else if (ch & 0x80)
	    tty_printf("\\%03o ", (unsigned char) ch);
	else
	    tty_printf("%c ", ch);
    }
}

/*
 * return a static pointer to escape sequence made printable, for use in
 * definition-files
 */
char *seq_name(seq, len)
char *seq; int len;
{
    static char buf[LINELEN];
    char *p = buf;
    char c;
    /*
     * rules: control chars are written as ^X, where
     * X is char | 64
     * 
     * GH: codes > 0x80  ==>  octal \012
     *
     * special case: 0x7f is written ^?
     */
    while (len--) {
	c = *seq++;
        if (c == '^' || (c && strchr(SPECIAL_CHARS, c)))
	    *(p++) = ESC;
	
	if (c >= 0 && c < ' ') {
            *(p++) = '^';
            *(p++) = c | '@';
	} else if (c == 0x7f) {
	    *(p++) = '^';
	    *(p++) = '?';
        } else if (c & 0x80) {
	    /* GH: save chars with high bit set in octal */
	    sprintf(p, "\\%03o", (int)(unsigned char) c);
	    p += strlen(p);
	} else
	    *(p++) = c;
    }
    *p = '\0';
    return buf;
}

/*
 * read a single escape sequence from the keyboard
 * prompting user for it; return static pointer
 */
static char *read_seq(name, len)
char *name; int *len;
{	
    static char seq[CAPLEN];
    int i = 1, tmp;
    
    PRINT_INT ("#please press the key `%s' : ", name);
    tty_flush();

    if ((tmp = get_one_char(0)) >= 0)
	seq[0] = tmp;
    else {
	tty_puts("#unable to get key. Giving up.\n");
	return NULL;
    }
    
    while (i < CAPLEN - 1 &&
	   (tmp = get_one_char(uSEC_PER_mSEC * KBD_TIMEOUT)) >= 0)
	seq[i++] = tmp;
    *len = i;
    print_seq(seq, i);
    
    tty_putc('\n');
    if (seq[0] >= ' ' && seq[0] <= '~') {
	PRINT_INT ("#that is not a redefinable key.\n");
	return NULL;
    }
    return seq;
}

/*
 * show full definition of one binding,
 * with custom message
 */
static void show_single_bind(msg, p)
char *msg;
keynode *p;
{
    PRINT_INT ("#%s %s %s=%s\n", msg, p->name,
	       seq_name(p->sequence, p->seqlen),
	       p->funct == key_run_command ? p->call_data :
	       internal_functions[lookup_edit_function(p->funct)].name);
}

/*
 * interactively create a new keybinding
 */
static void define_new_key(name, command)
char *name, *command;
{
    char *seq;
    keynode *p;
    int seqlen, function;
    
    seq = read_seq(name, &seqlen);
    if (!seq) return;
    
    function = lookup_edit_name(command);
    
    for (p = keydefs; p; p = p->next)
	/* GH: don't allow binding of supersets of another bind */
	if (!bcmp(p->sequence, seq, MIN2(p->seqlen, seqlen))) {
	    show_single_bind("key already bound as:", p);
	    return;
	}
    if (function)
	add_keynode(name, seq, seqlen, internal_functions[function].funct, 
		    NULL, 0);
    else
	add_keynode(name, seq, seqlen, key_run_command, command, 0);
    
    if (echo_int) {
	tty_printf ("#new key binding: %s %s=%s\n", name, 
		    seq_name(seq, seqlen), command);
    }
}

/*
 * create/list/edit/delete bindings
 */
static void cmd_bind(arg)
char *arg;
{
    arg = skipspace(arg);
    if (!*arg)
	show_keybinds(0);
    else if (!strcmp(arg, "edit"))
	show_keybinds(1);
    else
	parse_bind(arg);
}

/*
 * change the escape sequence of an existing binding
 */
static void cmd_rebind(arg)
char *arg;
{
    char rawseq[CAPLEN], *seq, **old, *tmp;
    keynode **kp, *p;
    int seqlen;
    
    arg = skipspace(arg);
    if (!*arg) {
	PRINT_INT ("#rebind: missing key.\n");
	return;
    }
    
    seq = first_valid(arg, ' ');
    if (*seq) {
	*seq++ = '\0';
	seq = skipspace(seq);
    }
    
    kp = lookup_key(arg);
    if (!kp || !*kp) {
	PRINT_INT ("#no such key: %s\n", arg);
	return;
    }
    
    if (!*seq) {
	seq = read_seq(arg, &seqlen);
	if (!seq)
	    return;
    } else {
	(void)unescape_seq(rawseq, seq, &seqlen);
	seq = rawseq;
    }

    for (p = keydefs; p; p = p->next) {
	if (p == *kp)
	    continue;
	if (seqlen == p->seqlen && !bcmp(p->sequence, seq, seqlen)) {
	    show_single_bind("key already bound as:", p);
	    return;
	}
    }
    
    tmp = (char*)malloc((*kp)->seqlen = seqlen);
    if (!tmp) {
	errmsg("malloc");
	return;
    }
    old = &((*kp)->sequence);
    if (*old)
	free(*old);
    *old = tmp;
    bcopy (seq, *old, seqlen);

    if (echo_int)
	show_single_bind("redefined key:", *kp);
}

/* ARGSUSED */
static void cmd_rebindall(arg)
{
    keynode *kp;
    unsigned char *seq;
    
    for (kp = keydefs; kp; kp = kp->next) {
	seq = kp->sequence;
	if (kp->seqlen == 1 && seq[0] < ' ')
	    ;
	else if (kp->seqlen == 2 && seq[0] == '\033' && isalnum(seq[1]))
	    ;
	else {
	    cmd_rebind(kp->name);
	    if (error)
		break;
	}
    }
}

/* ARGSUSED */
static void cmd_rebindALL(arg)
{
    keynode *kp;
    
    for (kp = keydefs; kp; kp = kp->next) {
	cmd_rebind(kp->name);
	if (error)
	    break;
    }
}

/*
 * list keyboard bindings
 */
static void show_keybinds(edit)
char edit;
{
    keynode *p;
    int count = 0;
    for (p = keydefs; p; p = p->next) {
	if (edit != (p->funct == key_run_command)) {
	    if (!count) {
		if (edit) {
		    PRINT_INT ("#line-editing keys:\n");
		} else {
		    PRINT_INT ("#user-defined keys:\n");
		}
	    }
	    show_single_bind("bind", p);
	    ++count;
	}
    }
    if (!count) {
        PRINT_INT ("#no key bindings defined right now.\n");
    }
}

/*
 * parse the #bind command non-interactively.
 */
void parse_bind_noninteractive(arg)
char *arg;
{
    char rawseq[CAPLEN], *p, *seq;
    int function, seqlen;
    keynode **kp;
    
    p = strchr(arg, ' ');
    if (!p) {
        PRINT_INT ("#syntax error: #bind %s\n", arg);
        return;
    }
    *(p++) = '\0';
    seq = p = skipspace(p);
    
    p = unescape_seq(rawseq, p, &seqlen);
    if (!*p || !*(p+1)) {
        PRINT_INT ("#syntax error: #bind %s %s\n", arg, seq);
        return;
    }
    *p++ = '\0';
    
    kp = lookup_key(arg);
    if (kp && *kp)
	delete_keynode(kp);
    
    if ((function = lookup_edit_name(p)))
	add_keynode(arg, rawseq, seqlen, 
		    internal_functions[function].funct, NULL);
    else
	add_keynode(arg, rawseq, seqlen, key_run_command, p);
    
    if (echo_int) {
	PRINT_INT ("#%s: %s %s=%s\n", (kp && *kp) ?
		   "redefined key" : "new key binding", arg, seq, p);
    }
}

/*
 * parse the argument of the #bind command (interactive)
 */
static void parse_bind(arg)
char *arg;
{
    char *p, *q, *command;
    char *name = arg;
    keynode **np;
    int function;
    
    p = first_valid(arg, '=');
    q = first_valid(arg, ' ');
    q = skipspace(q);
    
    if (*p && *q && p > q) {	
        parse_bind_noninteractive(arg);
	return;
    }
    
    if (*p) {
        *(p++) = '\0';
        np = lookup_key(name);
        if (*p) {
            command = p;
            if (*np) {
                if ((*np)->funct == key_run_command)
		    free((*np)->call_data);
		if ((function = lookup_edit_name(command))) {
		    (*np)->call_data = NULL;
		    (*np)->funct = internal_functions[function].funct;
		} else {
                    (*np)->call_data = str_dup(command);
		    (*np)->funct = key_run_command;
		}
                if (echo_int) {
                    PRINT_INT ("#redefined key: %s %s=%s\n", name,
			       seq_name((*np)->sequence, (*np)->seqlen),
			       command);
                }
            } else
		define_new_key(name, command);
        } else {
            if (*np) {
                if (echo_int)
		    show_single_bind("deleting key binding:", *np);
		delete_keynode(np);
            } else {
                PRINT_INT ("#no such key: %s\n", name);
            }
        }
    } else {
        np = lookup_key(name);
        if (*np) {
	    char buf[LINELEN];
            sprintf(buf, "#bind %s %s=%s", name, 
		    seq_name((*np)->sequence, (*np)->seqlen),
		    (*np)->funct == key_run_command ? (*np)->call_data : 
		    internal_functions[lookup_edit_function((*np)->funct)].name);
	    to_input_line(buf);
	} else {
            PRINT_INT ("#no such key: %s\n", name);
        }
    }
}

static void cmd_map(arg)
char *arg;
{
    arg = skipspace(arg);
    if (!*arg)  /* show map */
	map_show();
    else if (*arg == '-')  /* retrace steps without walking */
	map_retrace(atoi(arg + 1), 0);
    else
	parse_walk(arg, 1, 1);
}

static void cmd_retrace(arg)
char *arg;
{
    map_retrace(-atoi(arg), 1);
}


static void cmd_alias(arg)
char *arg;
{
    if (!*arg)
	show_aliases();
    else
	parse_alias(arg);
}

static char *trivial_eval(arg, buf, name)
char *arg, *buf, *name;
{
    char *tmp = skipspace(arg);
    
    if (*tmp=='(') {
	arg = tmp + 1;
	tmp = buf;
	(void)eval(&tmp, &arg, PRINT_AS_TEXT);
	if (error) return NULL;
	if (*arg != ')') {
	    PRINT_INT ("#%s: ", name);
	    print_error(error=MISSING_PAREN_ERROR);
	    return NULL;
	}
	arg = buf;
    }
    else {
	if (*arg == ' ') arg++;
	unescape(arg);
    }
    
    return arg;
}

static void cmd_add(arg)
char *arg;
{
    char buf[LINELEN];
    
    arg = trivial_eval(arg, buf, "add");
    if (!error)
	while (*arg) {
	    arg = split_first_word(buf, arg);
	    if (strlen(buf) >= MIN_WORD_LEN)
		put_word(buf);
	}
}



static void cmd_snoop(arg)
char *arg;
{
    if (!*arg) {
	PRINT_INT ("#snoop: no connection name.\n");
    }
    else
	tcp_togglesnoop(arg);
}

/*ARGSUSED*/
static void cmd_stop(arg)
char *arg;
{
    delaynode *dying;
    
    while (delays) {
	dying = delays;
	delays = dying->next;
	dying->when.tv_sec = now.tv_sec;
	dying->when.tv_usec = now.tv_usec;
	dying->next = dead_delays;
	dead_delays = dying;
    }
    if (echo_int) {
	PRINT_INT ("#all delayed labels are now disabled.\n");
    }
}


/*
 * parse the #alias command
 */
void parse_alias(str)
char *str;
{
    char *left, *right;
    aliasnode **np;
    char buf[LINELEN];
    
    left = str = skipspace(str);
    
    str = first_regular(str, '=');
    
    if (*str == '=') {
        *str = '\0';
        right = ++str;
        unescape(left);
        np = lookup_alias(left);
        if (!*str) {
            /* delete alias */
            if (*np) {
                if (echo_int) {
                    PRINT_INT ("#deleting alias: %s=%s\n", left, (*np)->subst);
                }
                delete_aliasnode(np);
            } else {
                PRINT_INT ("#unknown alias, cannot delete: %s\n", left);
            }
        } else {
            /* add/redefine alias */
	    
	    /* direct recursion is supported (alias CAN be defined by itself) */
            if (*np) {
                free((*np)->subst);
                (*np)->subst = str_dup(right);
            } else {
                add_aliasnode(left, right);
            }
            if (echo_int) {
                PRINT_INT ("#new alias: %s=%s\n", left, right);
            }
        }
    } else {
        /* show alias */
	
        *str = '\0';
        unescape(left);
        np = lookup_alias(left);
        if (*np) {
            char *buf2=buf;
            sprintf(buf, "#alias ");
            while (*buf2) buf2++;
            escape_specials(left, buf2);
            while (*buf2) buf2++;
            *buf2++ = '=';
            strcpy(buf2, (*np)->subst);
	    if (!internal)
		internal = 1;
            to_input_line(buf);
        } else {
            PRINT_INT ("#unknown alias, cannot show: %s\n", left);
        }
    }
}

#define ONPROMPT (onprompt ? "prompt" : "action")

/*
 * parse the #action and #prompt commands
 * this function is too damn complex because of the hairy syntax. it should be
 * split up or rewritten as an fsm instead.
 */
void parse_action(str, onprompt)
char *str;
int onprompt;
{
    char *p, *label, *pattern, *command, buf[LINELEN], *tmp=buf;
    actionnode **np = NULL;
    char sign, assign, haspattern;
    char active, type = ACTION_WEAK, kind;
#ifdef USE_REGEXP
    regex_t *regexp = 0;
#else
    void *regexp = 0;
#endif
    
    sign = *(p = skipspace(str));
    if (!sign) {
	PRINT_INT ("%s: no arguments given\n", ONPROMPT);
	return;
    }
    
    label = str = p + 1;
    
    switch (sign) {
      case '+':
      case '-':		/* edit */
      case '<':		/* delete */
      case '=':		/* list */
	assign = sign;
	break;
#ifdef USE_REGEXP
      case '%': /* action_chars[ACTION_REGEXP] */
	type = ACTION_REGEXP;
	/* falltrough */
#endif
      case '>': /* action_chars[ACTION_WEAK] */
	
	/* define/edit */
	assign = '>';
	sign = *(p + 1);
	if (!sign) {
	    PRINT_INT ("#%s: label expected\n", ONPROMPT);
	    return;
	} else if (sign == '+' || sign == '-')
	    str++;
	else
	    sign = '+';
	break;
      default:
	assign = 0;	/* anonymous action */
	break;
    }
    
    /* labelled action: */
    if (assign != 0) {
        p = label = str;
        p = first_regular(p, ' ');
        if (*p) *p++ = '\0'; /* p points to start of pattern, or to \0 */
        unescape(label);
	
	if (!*label) {
	    PRINT_INT ("#%s: label expected\n", ONPROMPT);
	    return;
	}
	
	if (onprompt)
	    np = lookup_prompt(label);
	else
	    np = lookup_action(label);
	
	/* '<' : remove action */
        if (assign == '<') {
            if (!np || !*np) {
                PRINT_INT ("#no %s, cannot delete label: %s\n",
			   ONPROMPT, label);
            }
            else {
		if (onprompt)
		    delete_prompt(np);
		else
		    delete_action(np);
	    }
	    
	    /* '>' : define action */
        } else if (assign == '>') {
#ifndef USE_REGEXP
	    if (type == ACTION_REGEXP) {
		PRINT_INT ("#error: regexp not allowed\n");
		return;
	    }	    
#endif
	    
            if (sign == '+')
		active = 1;
	    else
		active = 0;
	    
	    if (!*label) {
		PRINT_INT ("#%s: label expected\n", ONPROMPT);
		return;
	    }
	    
            p = skipspace(p);
            if (*p == '(') {
		p++;
		kind = eval(&tmp, &p, PRINT_AS_TEXT);
		if (error) return;
		if (kind != TYPE_TXT) {
		    PRINT_INT ("#%s: ", ONPROMPT);
		    print_error(error=NO_STRING_ERROR);
		    return;
		}
		pattern = buf;
		if (*p) 
		    p = skipspace(++p);
		if ((haspattern = *p == '='))
		    p++;
	    }
            else {
		p = first_regular(pattern = p, '=');
		if ((haspattern = *p))
		    *p++ = '\0';
		unescape(pattern);
	    }

	    if (!*pattern) {
		PRINT_INT ("#error: pattern of #%ss must be non-empty.\n", ONPROMPT);
		return;
	    }

#ifdef USE_REGEXP
	    if (type == ACTION_REGEXP && haspattern) {
		int errcode;

		regexp = (regex_t *)malloc(sizeof(regex_t)); 
		if (!regexp) {
		    errmsg("malloc");
		    return;
		}
		
		if ((errcode = regcomp(regexp, pattern, REG_EXTENDED))) {
		    char *buf;
		    int n;

		    n = regerror(errcode, regexp, NULL, 0);
		    buf = (char *)malloc(n);
		    if (buf) {
			if (!regerror(errcode, regexp, buf, n))
			    errmsg("regerror");
			else {
			    PRINT_INT ("#error: regexp error %s\n", buf);
			}
			free(buf);
		    } else {
			error = NO_MEM_ERROR;
			errmsg("malloc");
		    }
		    regfree(regexp);
		    free(regexp);
		    return;
		}
	    }
#endif
            command = p;
	    
            if (haspattern) {
                if (np && *np) {
		    change_actionorprompt(*np, pattern, command, type, regexp, onprompt);
                    (*np)->active = active;
                } else {
		    if (onprompt)
		      add_new_prompt(label, pattern, command, active, 
				     type, regexp);
		    else
		      add_new_action(label, pattern, command, active, 
				     type, regexp);
		}		
            }
	    
	    /* '=': list action */
        } else if (assign == '='){
            if (np && *np) {
                char *buf2=buf;
                sprintf(buf, "#%s %c%c", ONPROMPT,
			action_chars[(*np)->type],
			(*np)->active ? '+' : '-');
                while (*buf2) buf2++;
                escape_specials(label, buf2);
                while (*buf2) buf2++;
                *(buf2++) = ' ';
                escape_specials((*np)->pattern, buf2);
                while (*buf2) buf2++;
                *buf2++ = '=';
                strcpy(buf2, (*np)->command);
                to_input_line(buf);
            } else {
                PRINT_INT ("#no %s, cannot list label: %s\n", ONPROMPT, label);
            }
	    
	    /* '+', '-': turn action on/off */
        } else {
            if (np && *np) {
                (*np)->active = (sign == '+');
                if (echo_int) {
                    PRINT_INT ("#%s %c%s %s is now o%s.\n", ONPROMPT,
			       action_chars[(*np)->type],
			       label,
			       (*np)->pattern, (sign == '+') ? "n" : "ff");
                }
            } else {
                PRINT_INT ("#no %s, cannot turn o%s label: %s\n", ONPROMPT,
			   (sign == '+') ? "n" : "ff", label);
            }
        }
	
	/* anonymous action  */
    } else {
	
	if (onprompt) {
	    PRINT_INT ("#anonymous prompts not supported.\n#please do yourself a favour and use labels for both actions and prompts.\n");
	    return;
	}
	
        pattern = str;
        command = first_regular(str, '=');
	
        if (*command == '=') {
            *(command++) = '\0';
	    unescape(pattern);
            np = lookup_action_pattern(pattern);
            if (*command)
		if (np && *np)
		change_actionorprompt(*np, NULL, command, ACTION_WEAK, NULL, 0);
	    else
		add_anonymous_action(pattern, command, ACTION_WEAK, NULL);
            else if (np && *np)
		delete_action(np);
            else {
                PRINT_INT ("#no action, cannot delete pattern: %s\n",
			   pattern);
                return;
            }
        } else {
	    unescape(pattern);
            np = lookup_action_pattern(pattern);
            if (np && *np) {
                char *buf2=buf;
                strcpy(buf, "#action ");
                while (*buf2) buf2++;
                escape_specials((*np)->pattern, buf2);
                while (*buf2) buf2++;
                *buf2++ = '=';
                strcpy(buf2, (*np)->command);
                to_input_line(buf);
            } else {
                PRINT_INT ("#no action, cannot show pattern: %s\n", pattern);
            }
        }
    }
}

/*
 * delete an action node
 */
static void delete_action(nodep)
actionnode **nodep;
{
    if (echo_int) {
        PRINT_INT ("#deleting action: >%c%s %s\n", (*nodep)->active ?
		   '+' : '-', (*nodep)->label, (*nodep)->pattern);
    }
    delete_actionnode(nodep);
}

/*
 * delete a prompt node
 */
static void delete_prompt(nodep)
actionnode **nodep;
{
    if (echo_int) {
        PRINT_INT ("#deleting prompt: >%c%s %s\n", (*nodep)->active ?
		   '+' : '-', (*nodep)->label, (*nodep)->pattern);
    }
    delete_promptnode(nodep);
}

/*
 * change fields of an existing action node
 * pattern or commands can be NULL if no change
 */
static void change_actionorprompt(node, pattern, command, type, q, onprompt)
actionnode *node; char *pattern, *command; void *q;
int type, onprompt;
{
#ifdef USE_REGEXP
    if (node->type == ACTION_REGEXP && node->regexp) {
	regfree(node->regexp);
	free(node->regexp);
    }
    node->regexp = (regex_t *)q;
#endif
    if (pattern) {
        free(node->pattern);
        node->pattern = str_dup(pattern);
	node->type = type;
    }
    if (command) {
        free(node->command);
        node->command = str_dup(command);
    }
    
    if (echo_int) {
        PRINT_INT ("#changed %s %c%c%s %s=%s\n", ONPROMPT,
		   action_chars[node->type],
		   node->active ? '+' : '-',
		   node->label, node->pattern, node->command);
    }
}

/*
 * add an action with numbered label
 */
void add_anonymous_action(pattern, command, type, q)
char *pattern, *command; void *q;
int type;
{
    static last = 0;
    char label[16];
    do {
        sprintf(label, "%d", ++last);
    } while (*lookup_action(label));
    add_new_action(label, pattern, command, 1, type, q);
}

/*
 * create new action
 */
static void add_new_action(label, pattern, command, active, type, q)
char *label, *pattern, *command; int active; void *q;
int type;
{
    add_actionnode(pattern, command, label, active, type, q);
    if (echo_int) {
        PRINT_INT ("#new action: %c%c%s %s=%s\n", 
		   action_chars[type],
		   active ? '+' : '-', label,
		   pattern, command);
    }
}

/*
 * create new prompt
 */
static void add_new_prompt(label, pattern, command, active, type, q)
char *label, *pattern, *command; int active; void *q;
int type;
{
    add_promptnode(pattern, command, label, active, type, q);
    if (echo_int) {
        PRINT_INT ("#new prompt: %c%c%s %s=%s\n", 
		   action_chars[type],
		   active ? '+' : '-', label,
		   pattern, command);
    }
}

/*
 * parse arguments to the #mark command
 */
void parse_marker(str)
char *str;
{
    char *p;
    marknode **np;
    
    if (*str == '=') {
	PRINT_INT ("#marker must be non-null.\n");
	return;
    }
    p = first_regular(str, '=');
    if (!*p) {
        unescape(str);
        np = lookup_marker(str);
        if (*np) {
            char buf[LINELEN], *buf2=buf;
            strcpy(buf, "#mark ");
            while (*buf2) buf2++;
            escape_specials((*np)->pattern, buf2);
            while (*buf2) buf2++;
            *buf2++ = '=';
            strcpy(buf2, attr_name((*np)->attrcode));
            to_input_line(buf);
        } else {
            PRINT_INT ("#unknown marker, cannot show: %s\n", str);
        }
	
    } else {
        int attrcode;
	char *p2;

        *(p++) = '\0';
	p = skipspace(p);
        unescape(str);
	p2 = str;
	while(*p2) {
	  if(ISMARKWILDCARD(*p2) &&
	     ISMARKWILDCARD(*(p2 + 1))) {
	    PRINT_INT ("#error: two wildcards (& or $) may not be next to eachother\n");
	    return;
	  }
	  p2++;
	}

        np = lookup_marker(str);
        attrcode = parse_attributes(p);
        if (attrcode == -1) {
            PRINT_INT ("#invalid attribute syntax.\n");
            if (echo_int) show_attr_syntax();
        } else if (!*p)
	    if (*np) {
		if (echo_int) {
		    PRINT_INT ("#deleting mark: %s=%s\n", (*np)->pattern,
			       attr_name((*np)->attrcode));
		}
		delete_marknode(np);
	    } else {
		PRINT_INT ("#unknown marker, cannot delete: %s\n", str);
	    }
        else {
            if (*np) {
                (*np)->attrcode = attrcode;
                if (echo_int) {
                    PRINT_INT ("#changed");
                }
            } else {
                add_marknode(str, attrcode);
                if (echo_int) {
                    PRINT_INT ("#new");
                }
            }
            if (echo_int)
		tty_printf(" mark: %s=%s\n", str, attr_name(attrcode));
        }
    }
}

/*
 * show defined marks
 */
static void show_marks()
{
    marknode *p;
    PRINT_INT ("#%s marker%s defined%c\n", markers ? "the following" : "no",
	       (markers && !markers->next) ? " is" : "s are",
	       markers ? ':' : '.');
    for (p = markers; p; p = p->next)
	tty_printf("#mark %s=%s\n", p->pattern, attr_name(p->attrcode));
}

/*
 * show defined aliases
 */
static void show_aliases()
{
    aliasnode *p;
    
    PRINT_INT ("#%s alias%s defined%c\n", aliases ? "the following" : "no",
	       (aliases && !aliases->next) ? " is" : "es are",
	       aliases ? ':' : '.');
    for (p = aliases; p; p = p->next)
	tty_printf("#alias %s=%s\n", p->name, p->subst);
}

/*
 * show defined actions
 */
static void show_actions()
{
    actionnode *p;
    
    PRINT_INT ("#%s action%s defined%c\n", actions ? "the following" : "no",
	       (actions && !actions->next) ? " is" : "s are", actions ? ':' : '.');
    for (p = actions; p; p = p->next)
	tty_printf("#action %c%c%s %s=%s\n",
		   action_chars[p->type],
		   p->active ? '+' : '-', p->label,
		   p->pattern, p->command);
}

/*
 * show defined prompts
 */
static void show_prompts()
{
    promptnode *p;
    
    PRINT_INT ("#%s prompt%s defined%c\n", prompts ? "the following" : "no",
	       (prompts && !prompts->next) ? " is" : "s are", prompts ? ':' : '.');
    for (p = prompts; p; p = p->next)
	tty_printf("#prompt %c%c%s %s=%s\n",
		   action_chars[p->type],
		   p->active ? '+' : '-', p->label,
		   p->pattern, p->command);
}

/*
 * add escapes to protect special characters from being escaped.
 */
void escape_specials(str, p)
char *str, *p;
{
    while (*str) {
	if (!(*str >= ' ' && *str < 127)) {
	    sprintf (p, "\\%03o", (int)(unsigned char) *str++);
	    p += strlen(p);
	} else {
	    if (*str == ESC) {
		while (*str == ESC)
		    *p++ = *str++;
	    
		if (!*str || *str == ESC2)
		    *p++ = ESC2;
		else
		    *p++ = ESC;
	    }      
	    else if (strchr(SPECIAL_CHARS, *str))
		*p++ = ESC;
	
	    if (*str)
		*p++ = *str++;
	}
    }
    *p = '\0';
}

static char *redirect(arg, buf, kind, name, also_num, start, end)
char *arg, *buf, *kind, *name, also_num;
long *start, *end;
{
    char *tmp = skipspace(arg), k;
    char buf2[LINELEN];
    int type, i;
    
    k = *tmp;
    if (k == '!' || k == '<')
	arg = ++tmp;
    else
	k = 0;
    
    *start = *end = 0;
    
    if (*tmp=='(') {
	arg = tmp + 1;
	tmp = buf;
	type = eval(&tmp, &arg, PRINT_AS_TEXT);
	if (error)
	    return NULL;
	if (type!=TYPE_TXT && !also_num) {
	    PRINT_INT ("#%s: ", name);
	    print_error(error=NO_STRING_ERROR);
	    return NULL;
	}
	for (i=0; i<2; i++) if (*arg == CMDSEP) {
	    arg++;
	    if (!i && *arg == CMDSEP) {
		*start = 1;
		continue;
	    }
	    else if (i && *arg == ')') {
		*end = LONG_MAX;
		continue;
	    }
	    
	    tmp = buf2;
	    type = eval(&tmp, &arg, PRINT_AS_NUM);
	    if (error)
		return NULL;
	    if (type != TYPE_NUM) {
		PRINT_INT ("#%s: ", name);
		print_error(error=NO_NUM_VALUE_ERROR);
		return NULL;
	    }
	    if (i)
		*end = *(long*)buf2;
	    else
		*start = *(long*)buf2;
	}
	if (*arg != ')') {
	    PRINT_INT ("#%s: ", name);
	    print_error(error=MISSING_PAREN_ERROR);
	    return NULL;
	}
	arg = buf;
	if (!*start && *end)
	    *start = 1;
    }
    else
	unescape(arg);
    
    *kind = k;
    return arg;
}

static void cmd_print(arg)
char *arg;
{
    /*     FILE *pclose(); */
    
    char buf[LINELEN], kind;
    long start, end, i = 1;
    FILE *fp;
    
    tty_clear_input_line(opt_compact);
    
    if (!internal)
	internal = 1;
    
    if (!*arg) {
	smart_print(var[NUMVAR].str ? var[NUMVAR].str : "");
	return;
    }
    
    arg = redirect(arg, buf, &kind, "print", 1, &start, &end);
    if (error || !arg) return;
    
    if (kind) {
	fp = (kind == '!') ? popen(arg, "r") : fopen(arg, "r");
	if (!fp) {
	    PRINT_INT ("#print: #error opening %s\n", arg);
	    error = SYNTAX_ERROR;
	    return;
	}
	while (!error && (!start || i <= end) && fgets(buf, LINELEN, fp))
	    if (!start || i++>=start)
	    tty_puts(buf);
	tty_putc('\n');
	
	if (kind == '!') pclose(fp); else fclose(fp);
    } else
	smart_print(arg);
}

static void cmd_var(arg)
char *arg;
{
    /*     FILE *pclose(); */
    
    char buf[LINELEN], *tmp, *tmp2, kind, type, right = 0, deleting = 0;
    varnode **ptr_named_var = NULL, *named_var = NULL;
    FILE *fp;
    long start, end, i = 1;
    int len, index;
    
    arg = skipspace(arg);
    tmp = first_regular(arg, '=');
    
    if (*tmp) {
	*tmp++ = '\0';     /* skip the = */
	right = 1;
	if (!*tmp)
	    deleting = 1;
	else if (*tmp == ' ')
	    tmp++;
    }
    
    if (*arg == '$')
	type = TYPE_TXT_VAR;
    else if (*arg == '@')
	type = TYPE_NUM_VAR;
    else {
	varnode *v;
	int found = 0, type;
	char buf[LINELEN];

	for(type = 0; type < 2; type++) {
	    v = named_vars[type];
	    while (v) {
		if (!found) {
		    /*
		     * there are always at least the two permanent variables
		     * $prompt and $last_line
		     */
		    PRINT_INT ("#the following named variables are defined:\n");
		    found = 1;
		}
		if (type == 0)
		    sprintf(buf, "%ld", v->num);
		else {
		    char *p;

		    escape_specials (v->str, &buf[1]);
		    buf[0] = '\"';
		    p = strchr(buf, 0);
		    *p++ = '\"';
		    *p = 0;
		}
		PRINT_INT("#(%c%s = %s)\n", type ? '$' : '@', v->name, buf);
		v = v->next;
	    }
	}
	return;
    }
    kind = *++arg;
    if (isalpha(kind) || kind == '_') {
	kind = type==TYPE_TXT_VAR ? 1 : 0;
	ptr_named_var = lookup_varnode(arg, kind);
	if (!*ptr_named_var) {
	    if (!deleting) {
		named_var = add_varnode(arg, kind);
		if (error)
		    return;
		if (echo_int) {
		    PRINT_INT ("#new variable: %s\n", arg - 1);
		}
	    } else {
		print_error(error=UNDEFINED_VARIABLE_ERROR);
		return;
	    }
	} else
	    named_var = *ptr_named_var;
	
	index = named_var->index;
    } else {
	kind = type==TYPE_TXT_VAR ? 1 : 0;
	
	tmp2=skipspace(arg);
	if (*tmp2 == '(') {
	    arg = tmp2+1;
	    tmp2 = buf;
	    index = eval(&tmp2, &arg, PRINT_AS_TEXT);
	    if (error) return;
	    if (index != TYPE_TXT) {
		print_error(error=NO_STRING_ERROR);
		return;
	    }
	    if (isdigit(*buf)) {
		index = atoi(buf);
		if (index < -NUMVAR || index >= NUMPARAM) {
		    print_error(error=OUT_RANGE_ERROR);
		    return;
		}
	    } else {
		varnode *named_var;
		
		if (!(named_var = *lookup_varnode(buf, kind))) {
		    if (!deleting) {
			named_var = add_varnode(buf, kind);
			if (error)
			    return;
			if (echo_int) {
			    PRINT_INT ("#new variable: %c%s\n", kind
				       ? '$' : '@', buf);
			}
		    } else {
			print_error(error=UNDEFINED_VARIABLE_ERROR);
			return;
		    }
		}
		index = named_var->index;
	    }
	} else {
	    /* an unnamed var */
	    tmp2 = buf;
	    index = eval(&tmp2, &arg, PRINT_AS_NUM);
	    if (error) return;
	    if (index != TYPE_NUM) {
		print_error(error=NO_NUM_VALUE_ERROR);
		return;
	    }
	    index = *(long *)buf;
	    if (index < -NUMVAR || index >= NUMPARAM) {
		print_error(error=OUT_RANGE_ERROR);
		return;
	    }
	    /* ok, it's an unnamed var */
	}
    }

    if (type == TYPE_TXT_VAR && right && !deleting && !var[index+NUMVAR].str) {
	if ((tmp2 = malloc(PARAMLEN)))
	    var[index+NUMVAR].str = tmp2;
	else {
	    print_error(error=NO_MEM_ERROR);
	    return;
	}
    }

    if (deleting) {
	if (*ptr_named_var) {
	    if (is_permanent_variable(*ptr_named_var)) {
		PRINT_INT ("#cannot delete variable: %s\n", arg - 1);
	    } else {
		delete_varnode(ptr_named_var, kind);
		if (echo_int) {
		    PRINT_INT ("#deleted variable: %s\n", arg - 1);
		}
	    }
	}
	else if ((type = TYPE_TXT_VAR)) {
	    if ((tmp2 = var[index+NUMPARAM].str))
		if (index < 0)
		free (tmp2);
	    else
		*tmp2 = '\0';
	}
	else
	    var[index+NUMPARAM].num = 0;
	
	return;
    }
    else if (!right) {
	tmp = buf;
	*tmp++ = '#';
	*tmp++ = '(';
	*tmp++ = type == TYPE_TXT_VAR ? '$' : '@';
	if (named_var)
	    sprintf(tmp, "%s=", named_var->name);
	else
	    sprintf(tmp, "%d=", index);
	while (*tmp)
	    tmp++;
	if (type == TYPE_TXT_VAR) {
	    *tmp++ = '\"';
	    escape_specials(var[index+NUMVAR].str, tmp);
	    while (*tmp)
		tmp++;
	    *tmp++ = '\"';
	}
	else {
	    sprintf(tmp, "%ld", *var[index+NUMVAR].num);
	    while (*tmp)
		tmp++;
	}
	*tmp++ = ')';
	*tmp = '\0';
	to_input_line(buf);
	return;
    }
    
    arg = redirect(tmp, buf, &kind, "var", 1, &start, &end);
    if (error || !arg) return;
    
    if (kind) {
	fp = (kind == '!') ? popen(arg, "r") : fopen(arg, "r");
	if (!fp) {
	    PRINT_INT ("#var: #error opening `%s'\n", arg);
	    error = SYNTAX_ERROR;
	    return;
	}
	len = 0;
	i = 1;
	while (!error && (!start || i<=end) && fgets(buf+len, LINELEN-len, fp))
	    if (!start || i++>=start)
	    len += strlen(buf + len);
	
	if (kind == '!') pclose(fp); else fclose(fp);
	if (len>PARAMLEN)
	    len = PARAMLEN;
	buf[len] = '\0';
	arg = buf;
    }
    
    if (type == TYPE_NUM_VAR) {
	arg = skipspace(arg);
	type = 1;
	len = 0;
	
	if (*arg == '-')
	    arg++, type = -1;
	else if (*arg == '+')
	    arg++;
	
	if (isdigit(kind=*arg)) while (isdigit(kind)) {
	    len*=10;
	    len+=(kind-'0');
	    kind=*++arg;
	}
	else {
	    PRINT_INT ("#var: ");
	    print_error (error=NO_NUM_VALUE_ERROR);
	    return;
	}
	*var[index+NUMVAR].num = len * type;
    }
    else
	strcpy(var[index+NUMVAR].str, arg);
}

static void cmd_send(arg)
char *arg;
{
    /*     FILE *pclose(); */
    
    char buf[LINELEN], kind;
    long start, end, i = 1;
    FILE *fp;
    
    arg = redirect(arg, buf, &kind, "send", 0, &start, &end);
    if (error || !arg) return;
    
    if (kind) {
	fp = (kind == '!') ? popen(arg, "r") : fopen(arg, "r");
	if (!fp) {
	    PRINT_INT ("#send: #error opening `%s'\n", arg);
	    error = SYNTAX_ERROR;
	    return;
	}
	while (!error && (!start || i<=end) && fgets(buf, LINELEN, fp))
	    if (!start || i++>=start) {
		buf[strlen(buf)-1] = '\0';
		if (echo_ext) {
		    PRINT_INT ("[%s]\n", buf);
		}
		tcp_write(buf);
	    }
	
	if (kind == '!') pclose(fp); else fclose(fp);
    } else {
	if (echo_ext) {
	    PRINT_INT ("[%s]\n", arg);
	}
	tcp_write(arg);
    }
}

static void cmd_write(arg)
char *arg;
{
    /*     FILE *pclose(); */
    
    char buf[LINELEN], buf2[LINELEN];
    char *tmp = skipspace(arg), kind;
    FILE *fp;
    
    kind = *tmp;
    if (kind == '!' || kind == '>')
	arg = ++tmp;
    else
	kind = 0;
    
    if (*tmp=='(') {
	arg = tmp + 1;
	tmp = buf;
	(void)eval(&tmp, &arg, PRINT_AS_TEXT);
	if (error)
	    return;
	
	if (*arg == CMDSEP) {
	    arg++;
	    tmp = buf2;
	    (void)eval(&tmp, &arg, PRINT_AS_TEXT);
	    if (error)
		return;
	}
	else {
	    PRINT_INT ("#write: ");
	    print_error(error=SYNTAX_ERROR);
	    return;
	}
	if (*arg != ')') {
	    PRINT_INT ("#write: ");
	    print_error(error=MISSING_PAREN_ERROR);
	    return;
	}
	arg = buf2;
	
	fp = (kind == '!') ? popen(arg, "w") : fopen(arg, kind ? "w" : "a");
	if (!fp) {
	    PRINT_INT ("#write: #error opening %s\n", arg);
	    error = SYNTAX_ERROR;
	    return;
	}
	fprintf(fp, "%s\n", buf);
	fflush(fp);
	if (kind == '!') pclose(fp); else fclose(fp);
    }
    else {
	PRINT_INT ("#write: ");
	print_error(error=MISMATCH_PAREN_ERROR);
	return;
    }
}

static void cmd_settimer(arg)
char *arg;
{
    char buf[LINELEN], *temp = buf;
    int type;
    long millisec;
    
    arg = skipspace(arg);
    if (*arg == '(') {
	arg++;
	type = eval(&temp, &arg, PRINT_AS_NUM);
	if (error) return;
	
	if (*arg == ')') {
	    if (type != TYPE_NUM) {
		PRINT_INT ("#settimer: ");
		print_error(error=NO_NUM_VALUE_ERROR);
	    }
	    else {
		vtime t;
		
		millisec = - *(long *)buf;
		t.tv_usec = (millisec % mSEC_PER_SEC) * uSEC_PER_mSEC;
		t.tv_sec  =  millisec / mSEC_PER_SEC;
		ref_time.tv_usec = now.tv_usec;
		ref_time.tv_sec  = now.tv_sec;
		add_vtime(&ref_time, &t);
	    }
	} else {
	    PRINT_INT ("#settimer: ");
	    print_error(error=MISSING_PAREN_ERROR);
	    return;
	}
    } else {
	PRINT_INT ("#settimer: ");
	print_error(error=MISMATCH_PAREN_ERROR);
    }
}

static void cmd_exe(arg)
char *arg;
{
    /*     FILE *pclose(); */
    
    char buf[LINELEN], kind;
    long start, end, i = 1;
    FILE *fp;
    
    arg = redirect(arg, buf, &kind, "exe", 0, &start, &end);
    if (error || !arg) return;
    
    if (kind) {
	fp = (kind == '!') ? popen(arg, "r") : fopen(arg, "r");
	if (!fp) {
	    PRINT_INT ("#exe: #error opening `%s'\n", arg);
	    error = SYNTAX_ERROR;
	    return;
	}
	while (!error && (!start || i<=end) && fgets(buf, LINELEN, fp))
	    if (!start || i++>=start) {
		buf[strlen(buf)-1] = '\0';
		parse_user_input(buf, 0);
	    }
	
	if (kind == '!') pclose(fp); else fclose(fp);
    } else
	parse_user_input(arg, 0);
}


static void cmd_emulate(arg)
char *arg;
{
    /*     FILE *pclose(); */
    
    char buf[LINELEN], kind;
    FILE *fp;
    long start, end, i = 1;
    int len;
    
    arg = redirect(arg, buf, &kind, "emulate", 0, &start, &end);
    if (error || !arg) return;
    
    if (kind) {
	fp = (kind == '!') ? popen(arg, "r") : fopen(arg, "r");
	if (!fp) {
	    PRINT_INT ("#emulate: #error opening `%s'\n", arg);
	    error = SYNTAX_ERROR;
	    return;
	}
	internal = -2;
	col0 = promptlen = promptstr[0] = surely_isprompt = '\0';
	
	while (!error && (!start || i<=end) && fgets(buf, LINELEN, fp))
	    if (!start || i++>=start)
	    process_remote_input(buf, strlen(buf));
	
	if (kind == '!') pclose(fp); else fclose(fp);
    } else {
	internal = -2;
	arg[len = strlen(arg)] = '\n';
	arg[++len] = '\0';
	col0 = promptlen = promptstr[0] = surely_isprompt = '\0';
	process_remote_input(arg, len);
    }
}

static void cmd_eval(arg)
char *arg;
{
    arg = skipspace(arg);
    if (*arg=='(') {
	arg++;
	(void)eval(NULL, &arg, PRINT_NOTHING);
	if (*arg != ')') {
	    PRINT_INT ("#(): ");
	    print_error(error=MISSING_PAREN_ERROR);
	}
    }
    else {
	PRINT_INT ("#(): ");
	print_error(error=MISMATCH_PAREN_ERROR);
    }
}

static void cmd_action(arg)
char *arg;
{
    if (!*arg)
	show_actions();
    else
	parse_action(arg, 0);
}

static void cmd_prompt(arg)
char *arg;
{
    if (!*arg)
	show_prompts();
    else
	parse_action(arg, 1);
}

/*ARGSUSED*/
static void cmd_isprompt(arg)
char *arg;
{
    if (tcp_fd == tcp_main_fd)
      surely_isprompt = 1;
}

static void cmd_capture(arg)
char *arg;
{
    arg = skipspace(arg);
    
    if (!*arg) {
        if (capfile) {
            fclose(capfile);
            capfile = NULL;
            if (echo_int) {
		PRINT_INT ("#end of capture to file.\n");
            }
        } else {
            PRINT_INT ("#capture to what file?\n");
        }
    } else {
        if (capfile) {
            PRINT_INT ("#capture already active.\n");
        } else {
            if ((capfile = fopen(arg, "w")) == NULL) {
                PRINT_INT ("#error writing file %s\n", arg);
            } else if (echo_int) {
                PRINT_INT ("#capture to `%s' active, `#capture' ends.\n", arg);
            }
        }
    }
}

static void cmd_record(arg)
char *arg;
{
    arg = skipspace(arg);
    
    if (!*arg) {
        if (recordfile) {
            fclose(recordfile);
            recordfile = NULL;
            if (echo_int) {
		PRINT_INT ("#end of record to file.\n");
            }
        } else {
            PRINT_INT ("#record to what file?\n");
        }
    } else {
        if (capfile) {
            PRINT_INT ("#record already active.\n");
        } else {
            if ((recordfile = fopen(arg, "w")) == NULL) {
                PRINT_INT ("#error writing file %s\n", arg);
            } else if (echo_int) {
                PRINT_INT ("#record to `%s' active, `#record' ends.\n", arg);
            }
        }
    }
}

static void cmd_prefix(arg)
char *arg;
{
    strcpy(prefixstr, arg);
    if (echo_int) {
	PRINT_INT ("#prefix %s.\n", *arg ? "set" : "cleared");
    }
}

#ifndef NO_SHELL
static void cmd_shell(arg)
char *arg;
{
    if (!*arg) {
        if (echo_int) {
	    PRINT_INT ("#that's easy.\n");
        }
    } else {
        tty_quit();
	
        system(arg);
	
        tty_init();
        tty_gotoxy(col0 = 0, line0 = lines -1);
        tty_puts(tty_clreoln);
    }
}
#endif

static void cmd_identify(arg)
char *arg;
{
    char id_string[10];
    
    strcpy(id_string, "~$#EI");
    
    edit_start[0] = edit_end[0] = '\0';
    if (*arg) {
        char *p = strchr(arg, ' ');
        if (p) {
            *(p++) = '\0';
            strcpy(edit_end, p);
        }
        strcpy(edit_start, arg);
    }
    tcp_write(id_string);
    internal = 0;
    if (echo_int) {
        PRINT_INT ("#identify: %s done!\n", id_string);
    }
    identified = 1;
}

/*ARGSUSED*/
static void cmd_edit(arg)
char *arg;
{
    editsess *sp;
    
    if (edit_sess) {
        for (sp = edit_sess; sp; sp = sp->next) {
	    PRINT_INT("# %s (%d)\n", sp->descr, sp->key);
	}
    } else {
	PRINT_INT("#no active editors.\n");
    }
}

static void cmd_cancel(arg)
char *arg;
{
    editsess *sp;
    
    if (!edit_sess) {
        PRINT_INT ("#no editing sessions to cancel.\n");
    } else {
        if (*arg) {
            for (sp = edit_sess; sp; sp = sp->next)
		if (atoi(arg) == sp->key) {
		    cancel_edit(sp);
		    break;
		}
            if (!sp) {
                PRINT_INT ("#no such editing session - use the session number.\n");
            }
        } else {
            if (edit_sess->next) {
                PRINT_INT ("#several editing sessions active, use #cancel <number>\n");
            } else
		cancel_edit(edit_sess);
        }
    }
}

static void cmd_help(arg)
char *arg;
{
    int i, size;
    char *text, *tmp;
    
    arg = skipspace(arg);
    if (*arg == '#') arg++;
    if (!*arg) {
        for (size = 50, i = 0; commands[i].name; i++)
	    size += strlen(commands[i].name) + strlen(commands[i].help) + 5;
	
	text = tmp = (char*)malloc(size);
	if (!text) {
	    errmsg("malloc");
	    return;
	}
	
        sprintf (tmp, "#help\n#commands available:\n");
	tmp += strlen(tmp);
	
        for (i = 0; commands[i].name; i++) {
	    sprintf(tmp, "#%s %s\n", commands[i].name, commands[i].help);
	    tmp += strlen(tmp);
	}
	
	if (!internal) internal = 1;
	message_edit(text, strlen(text), 1, 1);
    } else {
	FILE *f = fopen(helpfile, "r");
	char line[LINELEN];
	int len;
	
	if (!f) {
	    PRINT_INT ("#can't open help file `%s'\n", helpfile);
	} else {
	    while ((tmp = fgets(line, LINELEN, f)) &&
		   (line[0] != '@' || strncmp(line + 1, arg, strlen(arg))))
		;
	    if (tmp) {
	        if (!internal) internal = 1;
		if (!(text = (char*)malloc(size = LINELEN))) {
		    errmsg("malloc");
		    return;
		}
		
		strcpy(text, line); /* safe, line is not longer than LINELEN */
		i = strlen(line);
		
		while (fgets(line, LINELEN, f) && line[0] == '@')
		    ;     /* allow multiple commands to share the same help */
		
		do {
		    if ((len = strlen(line)) > size - i)
			/* Not enough space in current buffer */
			
			if (!(tmp = (char*)malloc(size += LINELEN))) {
			    errmsg("malloc");
			    free(text);
			    return;
			}
		    else {
			strcpy(tmp, text);
			free(text);
			text = tmp;
		    }
		    strcpy(text + i, line);
		    i += len;
		    
		} while (fgets(line, LINELEN, f) && line[0] != '@');
		
		message_edit(text, strlen(text), 1, 1);
	    } else {
		PRINT_INT ("#no entry for `%s' in the help file.\n", arg);
	    }
	}
    }
}

/*ARGSUSED*/
static void cmd_time(arg)
char *arg;
{
    struct tm *s;
    char buf[LINELEN];
    
    s = localtime((time_t *)&now.tv_sec);
    (void)strftime(buf, LINELEN - 1, "%a,  %d %b %Y  %H:%M:%S", s);
    PRINT_INT ("#current time is %s\n", buf);
}

void show_delaynode(p, in_or_at)
delaynode *p; char in_or_at;
{
    long d = diff_vtime(&p->when, &now);
    struct tm *s;
    char buf[LINELEN];
    
    s = localtime((time_t *)&p->when.tv_sec); 
    /* s now points to a calendar struct */
    if (in_or_at) {
	char buf2[LINELEN];
	
	if (in_or_at == 2) {
	    (void)strftime(buf, LINELEN - 1, "%H%M%S", s); /* writes time in buf */
	    sprintf(buf2, "#at %s (%s) %s", p->name, buf, p->command);
	}
	else
	    sprintf(buf2, "#in %s (%ld) %s", p->name, d, p->command);
	
	to_input_line(buf2);
    }
    else {
	(void)strftime(buf, LINELEN - 1, "%H:%M:%S", s);
	PRINT_INT ("#at (%s) #in (%ld) `%s' %s\n", buf, d, p->name, p->command);
    }
}


static void show_delays()
{
    delaynode *p;
    int n = (delays ? delays->next ? 2 : 1 : 0) +
	(dead_delays ? dead_delays->next ? 2 : 1 : 0);
    
    PRINT_INT ("#%s delay label%s defined%c\n", n ? "the following" : "no",
	       n == 1 ? " is" : "s are", n ? ':' : '.');
    for (p = delays; p; p = p->next)
	show_delaynode(p, 0);
    for (p = dead_delays; p; p = p->next)
	show_delaynode(p, 0);
}


static void change_delaynode(p, command, millisec)
delaynode **p; char *command; long millisec;
{
    delaynode *m=*p;
    
    *p = m->next;
    m->when.tv_usec = (millisec % mSEC_PER_SEC) * uSEC_PER_mSEC;
    m->when.tv_sec  =  millisec / mSEC_PER_SEC;
    add_vtime(&m->when, &now);
    if (*command)
	if (strlen(command) > strlen(m->command)) {
	    free((void*)m->command);
	    m->command = str_dup(command);
	}
    else
	strcpy(m->command, command);
    if (millisec < 0)
	add_node((defnode*)m, (defnode**)&dead_delays, rev_time_sort);
    else
	add_node((defnode*)m, (defnode**)&delays, time_sort);
    if (echo_int) {
	PRINT_INT ("#changed ");
	show_delaynode(m, 0);
    }
}

static void new_delaynode(name, command, millisec)
char *name, *command; long millisec;
{
    vtime t;
    
    t.tv_usec = (millisec % mSEC_PER_SEC) * uSEC_PER_mSEC;
    t.tv_sec  =  millisec / mSEC_PER_SEC;
    add_vtime(&t, &now);
    add_delaynode(name, command, &t, millisec < 0);
}


static void cmd_in(arg)
char *arg;
{
    char buf[LINELEN], *buf2 = buf, *name;
    long millisec;
    int type;
    delaynode **p;
    
    arg = skipspace(arg);
    if (!*arg) {
	show_delays();
	return;
    }
    
    arg = first_regular(name = arg, ' ');
    if (*arg)
	*arg++ = 0;
    
    unescape(name);
    
    p = lookup_delay(name, 0);
    if (!*p)  p = lookup_delay(name, 1);
    
    if (!*arg && !*p) {
	PRINT_INT ("#unknown delay label, cannot show: %s\n", name);
	return;
    }
    if (!*arg) {
	show_delaynode(*p, 1);
	return;
    }
    if (*arg != '(') {
	PRINT_INT ("#in: ");
	print_error(error=MISMATCH_PAREN_ERROR);
	return;
    }
    arg++;    /* skip the '(' */
    
    type = eval(&buf2, &arg, PRINT_AS_NUM);
    if (error) return;
    if (type!=TYPE_NUM) {
	PRINT_INT ("#in :");
	print_error(error=NO_NUM_VALUE_ERROR);
	return;
    }
    
    if (*arg == ')') {          /* skip the ')' */
	if (*++arg == ' ')
	    arg++;
    }
    else {
	PRINT_INT ("#in: ");
	print_error(error=MISSING_PAREN_ERROR);
	return;
    }
    
    arg = skipspace(arg);
    millisec = *(long *)buf;
    if (*p && millisec)
	change_delaynode(p, arg, millisec);
    else if (!*p && millisec)
	if (*arg)
	new_delaynode(name, arg, millisec);
    else {
	PRINT_INT ("#cannot create delay label without a command.\n");
    }
    else if (*p && !millisec) {
	if (echo_int) {
	    PRINT_INT ("#deleting delay label: %s %s\n", name, (*p)->command);
	}
	delete_delaynode(p);
    }
    else {
	PRINT_INT ("#unknown delay label, cannot delete: %s\n", name);
    }
}

static void cmd_at(arg)
char *arg;
{
    char buf[LINELEN], *buf2 = buf, *name;
    char dayflag=0;
    struct tm *twhen;
    int num, hour, minute, second;
    delaynode **p;
    long millisec;
    
    arg = skipspace(arg);
    if (!*arg) {
	show_delays();
	return;
    }
    
    arg = first_regular(name = arg, ' ');
    if (*arg)
	*arg++ = 0;
    
    unescape(name);
    
    p = lookup_delay(name, 0);
    if (!*p)  p = lookup_delay(name, 1);
    
    if (!*arg && !*p) {
	PRINT_INT ("#unknown delay label, cannot show: %s\n", name);
	return;
    }
    if (!*arg) {
	show_delaynode(*p, 2);
	return;
    }
    if (*arg != '(') {
	PRINT_INT ("#in: ");
	print_error(error=MISMATCH_PAREN_ERROR);
	return;
    }
    arg++;    /* skip the '(' */
    
    (void)eval(&buf2, &arg, PRINT_AS_TEXT);
    if (error) return;
    
    if (*arg == ')') {        /* skip the ')' */
	if (*++arg == ' ')
	    arg++;
    }
    else {
	PRINT_INT ("#at: ");
	print_error(error=MISSING_PAREN_ERROR);
	return;
    }
    
    arg = skipspace(arg);
    twhen = localtime((time_t *)&now.tv_sec); 
    /* put current year, month, day in calendar struct */
    
    buf2 = skipspace(buf);  /* convert time-string into hour, minute, second */
    if (!*buf2 || !isdigit(*buf2)) {
	PRINT_INT ("#at: ");
	print_error(error=NO_NUM_VALUE_ERROR);
	return;
    }
    num = atoi(buf2);
    second = num % 100;
    minute = (num / 100) % 100;
    hour   = num / 10000;
    if (second>59 || minute>59 || hour>23) {
	PRINT_INT ("#at: #error: invalid time.\n");
	return;
    }
    if (hour < twhen->tm_hour || 
	(hour == twhen->tm_hour && 
	 (minute < twhen->tm_min || 
	  (minute == twhen->tm_min && second <= twhen->tm_sec)))) {
	dayflag = 1;
        /* it is NOT possible to define an #at refering to the past */
    }

    /* if you use a time smaller than the current, it refers to tomorrow */
    
    millisec = (hour - twhen->tm_hour) * 3600 + (minute - twhen->tm_min) * 60 + 
	second - twhen->tm_sec + (dayflag ? 24*60*60 : 0);
    millisec *= mSEC_PER_SEC; /* Comparing time with current calendar,
			       we finally got the delay */
    millisec -= now.tv_usec / uSEC_PER_mSEC;
    
    if (*p)
	change_delaynode(p, arg, millisec);
    else
	if (*arg)
	new_delaynode(name, arg, millisec);
    else {
	PRINT_INT ("#cannot create delay label without a command.\n");
    }
}

static void cmd_do(arg)
char *arg;
{
    char buf[LINELEN], *tmp = buf;
    int type;
    long result;
    
    arg = skipspace(arg);
    if (*arg != '(') {
	PRINT_INT ("#do: ");
	print_error(error=MISMATCH_PAREN_ERROR);
	return;
    }
    arg++;
    
    type = eval(&tmp, &arg, PRINT_AS_NUM);
    if (error) return;
    
    if (type != TYPE_NUM) {
	PRINT_INT ("#do: ");
	print_error(error=NO_NUM_VALUE_ERROR);
	return;
    }
    
    if (*arg == ')') {          /* skip the ')' */
	if (*++arg == ' ')
	    arg++;
    }
    else {
	PRINT_INT ("#do: ");
	print_error(error=MISSING_PAREN_ERROR);
	return;
    }
    
    result = *(long *)buf;
    if (tmp!=buf && result > 0)
	while (!error && result--)
	(void)parse_instruction(arg, 1, 0);
    else {
	PRINT_INT ("#do: bogus repeat count.\n");
    }
}

static void cmd_if(arg)
char *arg;
{
    char buf[LINELEN], *tmp=buf;
    int type;
    
    arg = skipspace(arg);
    if (*arg!='(') {
	PRINT_INT ("#if: ");
	print_error(error=MISMATCH_PAREN_ERROR);
	return;
    }
    arg++;  /* skip the '(' */
    
    type = eval(&tmp, &arg, PRINT_AS_NUM);
    if (error) return;
    
    if (type!=TYPE_NUM) {
	PRINT_INT ("#if: ");
	print_error(error=NO_NUM_VALUE_ERROR);
	return;
    }
    
    if (*arg == ')')  {              /* skip the ')' */
	if (*++arg == ' ')
	    arg++;
    }
    else {
	PRINT_INT ("#if: ");
	print_error(error=MISSING_PAREN_ERROR);
	return;
    }
    
    if (tmp!=buf && *(long *)buf)
	(void)parse_instruction(arg, 0, 0);
    else {
	arg = get_next_instr(arg);
	if (!strncmp(arg = skipspace(arg), "#else ", 6))
	    (void)parse_instruction(arg + 6, 0, 0);
    }
}

static void cmd_for(arg)
char *arg;
{
    int type = TYPE_NUM, loop=MAX_LOOP;
    char buf[LINELEN], *tmp;
    char *check, *increm = 0;
    params param;
    
    arg = skipspace(arg);
    if (*arg != '(') {
	PRINT_INT ("#for: ");
	print_error(error=MISMATCH_PAREN_ERROR);
	return;
    }
    set_param(&param);
    
    arg = skipspace(arg + 1);    /* skip the '(' */
    if (*arg != CMDSEP)
	(void)eval(NULL, &arg, PRINT_NOTHING);       /* execute <init> */
    
    check = arg + 1;
    
    if (error)
	;
    else if (*arg != CMDSEP) {
	PRINT_INT ("#for: ");
	print_error(error=MISSING_SEPARATOR_ERROR);
    }
    else while (!error && loop &&
		(*(tmp=buf) = '\0', increm=check,
		 (type = eval(&tmp, &increm, PRINT_AS_NUM)) == TYPE_NUM &&
		 !error && *increm == CMDSEP && *(long *)buf)) {
	
	tmp = first_regular(increm + 1, ')');
	if (*tmp)
	    (void)parse_instruction(tmp + 1, 1, 1);
	else {
	    PRINT_INT ("#for: ");
	    print_error(error=MISSING_PAREN_ERROR);
	}
	
	if (!error) {
	    tmp = increm + 1;
	    if (*tmp != ')')
		(void)eval(NULL, &tmp, PRINT_NOTHING);
	}
	
	loop--;
    }
    if (error)
	;
    else if (increm && *increm != CMDSEP) {
	PRINT_INT ("#for: ");
	print_error(error=MISSING_SEPARATOR_ERROR);
    }
    else if (!loop) {
	PRINT_INT ("#for: ");
	print_error(error=MAX_LOOP_ERROR);
    }
    else if (type != TYPE_NUM) {
	PRINT_INT ("#for: ");
	print_error(error=NO_NUM_VALUE_ERROR);
    }
    if (error!=DYN_STACK_UND_ERROR && error!=DYN_STACK_OV_ERROR)
	restore_param();
}

static void cmd_while(arg)
char *arg;
{
    int type, loop=MAX_LOOP;
    char buf[LINELEN], *tmp;
    char *check;
    params param;
    
    arg = skipspace(arg);
    if (!*arg) {
	PRINT_INT ("#while: ");
	print_error(error=MISMATCH_PAREN_ERROR);
	return;
    }
    set_param(&param);
    
    check = ++arg;   /* skip the '(' */
    while (!error && loop &&
	   (*(tmp=buf) = '\0', arg=check, 
	    (type = eval(&tmp, &arg, PRINT_AS_NUM)) == TYPE_NUM &&
	    !error && *arg == ')' && *(long *)buf)) {
	
	if (*(tmp = arg + 1) == ' ')          /* skip the ')' */
	    tmp++;
	if (*tmp)
	    (void)parse_instruction(tmp, 1, 1);
	loop--;
    }
    if (error)
	;
    else if (*arg != ')') {
	PRINT_INT ("#while: ");
	print_error(error=MISSING_PAREN_ERROR);
    }
    else if (!loop) {
	PRINT_INT ("#while: ");
	print_error(error=MAX_LOOP_ERROR);
    }
    else if (type != TYPE_NUM) {
	PRINT_INT ("#while: ");
	print_error(error=NO_NUM_VALUE_ERROR);
    }
    if (error!=DYN_STACK_UND_ERROR && error!=DYN_STACK_OV_ERROR)
	restore_param();
}

void show_stat() {
    cmd_net(NULL);
    cmd_cpu(NULL);
}

/*ARGSUSED*/
static void cmd_net(arg)
char *arg;
{
    PRINT_INT ("#received from host: %ld bytes, sent to host: %ld bytes.\n",
	       received, sent);
}


#ifndef CLOCKS_PER_SEC
#  define CLOCKS_PER_SEC uSEC_PER_SEC
#endif
/* hope it works.... */

/*ARGSUSED*/
static void cmd_cpu(arg)
char *arg;
{
    float f, l;
    vtime _now;
    gettimeofday(&_now, NULL);
    f = (float)((cpu_clock = clock()) - start_clock) / (float)CLOCKS_PER_SEC;
    l = (float)diff_vtime(&_now, &start_time) / (float)mSEC_PER_SEC;
    PRINT_INT ("#CPU time used: %.3f sec. (%.2f%%)\n", f,
	       (l > 0 && l > f) ? f * 100.0 / l : 100.0);
}

#ifdef BUG_TELNET
static void cmd_color(arg)
char *arg;
{
    int attrcode;
    
    arg = skipspace(arg);
    if (!*arg) {
        strcpy(tty_modenorm, tty_modenormbackup);
	tty_puts(tty_modenorm);	
	if (echo_int) {
	    PRINT_INT ("#standard color cleared.\n");
	}
	return;
    }
    
    attrcode = parse_attributes(arg);
    if (attrcode == -1) {
        PRINT_INT ("#invalid attribute syntax.\n");
	if (echo_int)
	    show_attr_syntax();
    } else {
        int bg = BACKGROUND(attrcode), fg = FOREGROUND(attrcode);
        if (fg >= COLORS || bg >= COLORS) {
	    PRINT_INT ("#please specify foreground and background colors.\n");
	} else {
	    sprintf(tty_modenorm, "\033[;%c%d;%s%dm",
		    fg<LOWCOLORS ? '3' : '9', fg % LOWCOLORS,
		    bg<LOWCOLORS ? "4" :"10", bg % LOWCOLORS);
	    tty_puts(tty_modenorm);
	    if (echo_int) {
	        PRINT_INT ("#standard colour set.\n");
	    }
        }
    }
}
#endif

static void cmd_lines(arg)
char *arg;
{
    int i=0;
    
    if (*arg && (i = atoi(arg)) > 0)
	lines = i;
    if (echo_int || !i) {
	PRINT_INT ("#lines: %d\n", lines);
    }
}

static void cmd_connect(arg)
char *arg;
{
#ifdef TERM
    PRINT_INT ("#connect: multiple connections not supported in term version.\n");
#else
    char *s1, *s2, *s3 = NULL, *s4 = NULL;
    int argc = 1;
    
    if (!*skipspace(arg))
      if (tcp_count || tcp_attachcount) {
	  tcp_show();
	  return;
      } else {
	  s1 = "main";
	  s2 = *initstr ? initstr : NULL;
      }
    else {
	s1 = strtok(arg, " ");
	s2 = strtok(NULL, " ");
	if (s2 && *s2) {
	    argc++;
	    s3 = strtok(NULL, " ");
	    if (s3 && *s3) {
		argc++;
		s4 = strtok(NULL, " ");
		if (s4 && *s4)
		    argc++;
	    }
	}
    }
    
    if (argc <= 2) {
	if (*hostname)
	    tcp_open(s1, s2, hostname, atoi(portnumber));
	else {
	    PRINT_INT ("#syntax: #connect session-id [[init-str] [hostname] [port]]\n");
	}
    } else if (argc == 3) {
	if (!*hostname) {
	    strcpy(hostname, s2);
	    strcpy(portnumber, s3);
	}
	tcp_open(s1, NULL, s2, atoi(s3));
    } else {
	if (!*hostname) {
	    strcpy(hostname, s3);
	    strcpy(portnumber, s4);
	}
	tcp_open(s1, s2, s3, atoi(s4));
    }
#endif /* TERM */
}


static void cmd_spawn(arg)
char *arg;
{
    char s[LINELEN];
    if (*(arg = skipspace(arg))) {
	arg = split_first_word(s, arg);
	if (*arg && *s) {
	    tcp_spawn(s, arg);
	    return;
	}
    }
    PRINT_INT ("#syntax: #spawn connect-id command\n");
}

static void cmd_zap(arg)
char *arg;
{
    if (!*arg) {
	PRINT_INT ("#zap: no connection name.\n");
    } else
	tcp_close(arg);
}

static void cmd_qui(arg)
char *arg;
{
    PRINT_INT ("#you have to write '#quit' - no less, to quit!\n");
}

static void cmd_quit(arg)
char *arg;
{
    if (*arg) { /* no skipspace() here! */
	PRINT_INT ("#quit: spurious argument?\n");
    } else
	exitpowwow();
}

char *full_options_string = "#option %cexit %chistory %cwords %ccompact %cdebug %cecho %cinfo %ckeyecho %cspeedwalk %cwrap %cautoprint%s";

static void cmd_option(arg)
char *arg;
{
    if (*(arg = skipspace(arg))) {
	int len, i, count = 0;
	char *str[] = { "exit", "history", "words", "compact", "debug", "echo",
			"info", "keyecho", "speedwalk", "wrap", "autoprint",
			0 };
	char *varptr[] = { &opt_exit, &opt_history, &opt_words, &opt_compact,
			   &opt_debug, &echo_ext, &echo_int, &echo_key,
			   &opt_speedwalk, &opt_wrap, &opt_autoprint, 0 };
	enum { MODE_ON, MODE_OFF, MODE_TOGGLE, MODE_REP } mode;
	char buf[LINELEN], buf2[LINELEN], *p, *var, c;
	strcpy(buf2, "#option");
	while ((arg = skipspace(split_first_word(buf, arg))), *buf) {
	    c = *(p = buf);
	    switch (c) {
	     case '=': mode = MODE_REP; p++; break;
	     case '+': mode = MODE_ON;  p++; break;
	     case '-': mode = MODE_OFF; p++; break;
	     default:  mode = MODE_TOGGLE;     break;
	    }
	    len = strlen(p);
	    var = 0;
	    count++;
	    for (i=0; str[i]; i++) {
		if (strncmp(str[i], p, len) == 0) {
		    var = varptr[i];
		    break;
		}
	    }
	    if (!str[i]) {
		if (strncmp("none", p, len) == 0)
		    continue;
		else {
		    PRINT_INT("#syntax: #option [[+|-|=]<name>]\t\twhere <name> is one of:\nexit history words compact debug echo info keyecho speedwalk wrap autoprint\n");
		    return;
		}
	    }
	    if (var) {
		switch (mode) {
		 case MODE_REP:
		    strcat(buf2, *var ? " +" : " -");
		    strcat(buf2, p);
		    break;
		 case MODE_ON:  *var  = 1; break;
		 case MODE_OFF: *var  = 0; break;
		 default:       *var ^= 1; break;
		}
		/* as above, but always print status if
		 * `#option info' alone was typed */
		if (mode != MODE_REP && !*arg && count==1 &&
		    (echo_int || (mode == MODE_TOGGLE && var==&echo_int))) {
		    PRINT_INT("#option %s is now o%s.\n", str[i],
			      *var ? "n" : "ff");
		}
	    }
	    if (strcmp(buf2, "#option"))
		to_input_line(buf2);
	}
    } else {
	char buf[LINELEN];
	sprintf (buf, full_options_string,
		 opt_exit  ? '+' : '-',
		 opt_history ? '+' : '-',
		 opt_words ? '+' : '-',
		 opt_compact ? '+' : '-',
		 opt_debug ? '+' : '-',
		 echo_ext  ? '+' : '-',
		 echo_int  ? '+' : '-',
		 echo_key  ? '+' : '-',
		 opt_speedwalk ? '+' : '-',
		 opt_wrap  ? '+' : '-',
		 opt_autoprint ? '+' : '-', "");
	to_input_line(buf);
    }
}

static void cmd_file(arg)
char *arg;
{
    arg = skipspace(arg);
    if (*arg == '=') {
	set_deffile(++arg);
	if (echo_int) {
	    if (*arg) {
		PRINT_INT ("#save-file set to: %s\n", deffile);
	    } else {
		PRINT_INT ("#save-file is now undefined.\n");
	    }
	}
    } else if (*deffile) {
	char buf[LINELEN];
	sprintf(buf, "#file =%s", deffile);
	to_input_line(buf);
    } else {			  
	PRINT_INT ("#save-file not defined.\n");
    }
}

static void cmd_save(arg)
char *arg;
{ 
    arg = skipspace(arg);
    if (*arg) {
	set_deffile(arg);
	if (echo_int) {
	    PRINT_INT ("#save-file set to: %s\n", deffile);
	}
    }
    else if (!*deffile) {
	PRINT_INT ("#save-file not defined.\n");
	return;
    }
    
    if (*deffile && save_settings() > 0 && echo_int) {
	PRINT_INT ("#settings saved to file.\n");
    }
}

static void cmd_load(arg)
char *arg;
{
    arg = skipspace(arg);
    if (*arg) {
	set_deffile(arg);
	if (echo_int) {
	    PRINT_INT ("#save-file set to: %s\n", deffile);
	}
    }
    else if (!*deffile) {
	PRINT_INT ("#save-file not defined.\n");
	return;
    }
    
    if (read_settings() > 0 && echo_int) {
	PRINT_INT ("#settings loaded from file.\n");
    }
}

static void cmd_put(arg)
char *arg;
{
    char buf[LINELEN];
    
    arg = trivial_eval(arg, buf, "put");
    if (!error && *arg)
	put_in_history(arg);
}



static void show_history(count)
int count;
{
    int i = curline;
    
    if (!count) count = lines - 1;
    if (count >= MAXHIST) count = MAXHIST - 1;
    i -= count;
    if (i < 0) i += MAXHIST;
    
    if (!internal) internal = 1;
    while (count) {
	if (hist[i])
	    tty_printf("#%2d: %s\n", count, hist[i]);
	count--;
	if (++i == MAXHIST) i= 0;
    }
}

static void exe_history(count)
int count;
{
    int i = curline;
    char buf[LINELEN];
    
    if (count >= MAXHIST) count = MAXHIST - 1;
    i -= count;
    if (i < 0) i += MAXHIST;
    if (hist[i]) {
	strcpy(buf, hist[i]);
	parse_user_input(buf, 0);
    }
}


static void cmd_history(arg)
char *arg;
{
    char *tmp = skipspace(arg), buf[LINELEN], *buf2 = buf;
    int num = 0;
    
    if (history_done) {
	print_error(error=NESTED_HISTORY_ERROR);
	return;
    }
    history_done = 1;
    
    if (*tmp == '(') {
	tmp++;
	num = eval(&buf2, &tmp, PRINT_AS_NUM);
	if (error)
	    return;
	if (num != TYPE_NUM) {
	    PRINT_INT ("#history: ");
	    print_error(error=NO_NUM_VALUE_ERROR);
	    return;
	}
	num = (int)*(long *)buf;
    }
    else
	num = atoi(tmp);
    
    if (num > 0)
	exe_history(num);  
    else
	show_history(-num);
}

/*ARGSUSED*/
static void cmd_ver(arg)
char *arg;
{
    printver();
}

static void cmd_quote(arg)
char *arg;
{
    arg = skipspace(arg);
    if (!*arg)
	verbatim ^= 1;
    else if (!strcmp(arg, "on"))
	verbatim = 1;
    else if (!strcmp(arg, "off"))
	verbatim = 0;
    if (echo_int) {
        PRINT_INT ("#%s mode.\n", verbatim ? "verbatim" : "normal");
    }
}

